#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import argparse
import base64
import getpass
import logging
from typing import Optional, Any, Dict, List, Union

from .enterprise_common import EnterpriseCommand
from .security_audit import SecurityAuditReportCommand
from .. import api, crypto, utils, vault, vault_extensions
from .base import GroupCommand, Command, dump_report_data, fields_to_titles, report_output_parser
from ..breachwatch import BreachWatch
from ..params import KeeperParams
from ..error import CommandError
from ..proto import breachwatch_pb2, client_pb2


breachwatch_list_parser = argparse.ArgumentParser(prog='breachwatch-list', parents=[report_output_parser])
breachwatch_list_parser.add_argument('--all', '-a', dest='all', action='store_true',
                                     help='Display all breached records (default is to show only first 30 records)')
breachwatch_list_parser.add_argument('--owned', '-o', dest='owned', action='store_true',
                                     help='Display only breached records owned by user (omits records shared to user)')
breachwatch_list_parser.add_argument('--numbered', '-n', action='store_true',
                                     help='Display records as a numbered list')

#breachwatch_list_parser.add_argument('--ignored', '-i', dest='ignored', action='store_true', help='Display ignored records')


breachwatch_password_parser = argparse.ArgumentParser(prog='breachwatch-password')
breachwatch_password_parser.add_argument('passwords', type=str, nargs='*', help='Password')

breachwatch_scan_parser = argparse.ArgumentParser(prog='breachwatch-scan')


breachwatch_ignore_parser = argparse.ArgumentParser(prog='breachwatch-ignore')
breachwatch_ignore_parser.add_argument('records', type=str, nargs='+', help='Record UID to ignore')

breachwatch_report_parser = argparse.ArgumentParser(prog='breachwatch-report', parents=[report_output_parser])


def register_commands(commands):
    commands['breachwatch'] = BreachWatchCommand()


def register_command_info(aliases, command_info):
    aliases['bw'] = 'breachwatch'
    command_info['breachwatch'] = 'Manager vault scan results or perform a one-time check'


class BreachWatchCommand(GroupCommand):
    def __init__(self):
        super(BreachWatchCommand, self).__init__()
        self.register_command('list', BreachWatchListCommand(), 'Displays a list of breached passwords.')
        self.register_command('ignore', BreachWatchIgnoreCommand(), 'Ignores a record marked as a breached password.')
        self.register_command('password', BreachWatchPasswordCommand(),
                              'Performs a quick check of a password against the BreachWatch database.')
        self.register_command('scan', BreachWatchScanCommand(), 'Generate enterprise security audit data for records not previously scanned.')
        report_desc = 'Run report on BreachWatch audit data across all enterprise vaults (must be an admin).'
        self.register_command('report', BreachWatchReportCommand(), report_desc)

        self.default_verb = 'list'

    def validate(self, params):  # type: (KeeperParams) -> None
        if not params.breach_watch and not params.msp_tree_key:
            raise CommandError('breachwatch',
                               'BreachWatch is not active. Please contact Keeper Support or visit the Web Vault at https://keepersecurity.com/vault')


class BreachWatchListCommand(Command):
    def get_parser(self):
        return breachwatch_list_parser

    def execute(self, params, **kwargs):   # type: (KeeperParams, ...) -> Union[None, List[Any]]
        fmt = kwargs.get('format', 'table')
        out = kwargs.get('output')
        table = []
        report_data = None
        for record, _ in BreachWatch.get_records_by_status(params, ['WEAK', 'BREACHED'], kwargs.get('owned')):
            row = [record.record_uid, record.title, vault_extensions.get_record_description(record)]
            table.append(row)

        if table:
            table.sort(key=lambda x: x[1].casefold())
            total = len(table)
            if not kwargs.get('all', False) and total > 32:
                table = table[:30]
            columns = ['record_uid', 'title', 'login']
            columns = fields_to_titles(columns) if fmt != 'json' else columns
            report_data = dump_report_data(table, columns, title='Detected High-Risk Password(s)', row_number=kwargs.get('numbered'), fmt=fmt, filename=out)
            if len(table) < total:
                logging.info('')
                logging.info('%d records skipped.', total - len(table))
        else:
            logging.info('No breached records detected')

        has_records_to_scan = any(params.breach_watch.get_records_to_scan(params))
        if has_records_to_scan:
            logging.info('Some passwords in your vault are missing security audit data.\n'
                         'Use "breachwatch scan" command to report security audit data for any unscanned passwords')
        return report_data

class BreachWatchPasswordCommand(Command):
    def get_parser(self):  # type: () -> Optional[argparse.ArgumentParser]
        return breachwatch_password_parser

    def execute(self, params, **kwargs):  # type: (KeeperParams, **any) -> any
        passwords = kwargs.get('passwords')
        echo_password = True
        if not passwords:
            echo_password = False
            passwords = []
            try:
                password = getpass.getpass(prompt='Password to Check: ', stream=None)
                if not password:
                    return
                passwords.append(password)
            except KeyboardInterrupt:
                print('')

        euids = []
        for result in params.breach_watch.scan_passwords(params, passwords):
            if result[1].euid:
                euids.append(result[1].euid)
            pwd = result[0] if echo_password else "".rjust(len(result[0]), "*")
            print(f'{pwd:>16s}: {"WEAK" if result[1].breachDetected else "GOOD" }')
        if euids:
            params.breach_watch.delete_euids(params, euids)


class BreachWatchScanCommand(Command):
    def get_parser(self):  # type: () -> Optional[argparse.ArgumentParser]
        return breachwatch_scan_parser

    def execute(self, params, **kwargs):  # type: (KeeperParams, Any) -> Any
        records = [x[0] for x in params.breach_watch.get_records_to_scan(params)]
        record_passwords = dict()    # type: Dict[str, str]
        for record in records:
            if isinstance(record, vault.PasswordRecord):
                if record.password:
                    record_passwords[record.record_uid] = record.password
            elif isinstance(record, vault.TypedRecord):
                password_field = record.get_typed_field('password')
                if password_field:
                    password = password_field.get_default_value(str)
                    if password:
                        record_passwords[record.record_uid] = password

        if len(record_passwords):
            euid_to_delete = []
            bw_requests = []
            all_passwords = set(record_passwords.values())
            scans = {x[0]: x[1] for x in params.breach_watch.scan_passwords(params, all_passwords)}
            for record_uid, record_password in record_passwords.items():
                if params.breach_watch_records:
                    if record_uid in params.breach_watch_records:
                        bwr = params.breach_watch_records[record_uid]
                        if 'data_unencrypted' in bwr:
                            passwords = bwr['data_unencrypted'].get('passwords', [])
                            for password in passwords:
                                euid = password.get('euid')
                                if euid:
                                    euid_to_delete.append(base64.b64decode(euid))
                if record_password in scans:
                    bwrq = breachwatch_pb2.BreachWatchRecordRequest()
                    bwrq.recordUid = utils.base64_url_decode(record_uid)
                    bwrq.breachWatchInfoType = breachwatch_pb2.RECORD
                    bwrq.updateUserWhoScanned = True
                    hash_status = scans[record_password]
                    bw_password = client_pb2.BWPassword()
                    bw_password.value = record_password
                    bw_password.status = client_pb2.WEAK if hash_status.breachDetected else client_pb2.GOOD
                    bw_password.euid = hash_status.euid
                    bw_data = client_pb2.BreachWatchData()
                    bw_data.passwords.append(bw_password)
                    data = bw_data.SerializeToString()
                    try:
                        record_key = params.record_cache[record_uid]['record_key_unencrypted']
                        bwrq.encryptedData = crypto.encrypt_aes_v2(data, record_key)
                    except:
                        continue
                    bw_requests.append(bwrq)
            while bw_requests:
                chunk = bw_requests[0:999]
                bw_requests = bw_requests[999:]
                rq = breachwatch_pb2.BreachWatchUpdateRequest()
                rq.breachWatchRecordRequest.extend(chunk)
                api.communicate_rest(params, rq, 'breachwatch/update_record_data',
                                     rs_type=breachwatch_pb2.BreachWatchUpdateResponse)
                params.sync_data = True
            if euid_to_delete:
                params.breach_watch.delete_euids(params, euid_to_delete)
        if not kwargs.get('suppress_no_op') or record_passwords:
            logging.info(f'Reported security audit data for {len(record_passwords)} records.')
            if record_passwords:
                api.sync_down(params)


class BreachWatchIgnoreCommand(Command):
    def get_parser(self):  # type: () -> Optional[argparse.ArgumentParser]
        return breachwatch_ignore_parser

    def execute(self, params, **kwargs):  # type: (KeeperParams, any) -> any
        if not params.record_cache:
            return
        if not params.breach_watch_records:
            return
        records = kwargs.get('records')
        if not records:
            return

        record_uids = set()
        for record_uid in records:
            if record_uid in record_uids:
                continue
            record_uids.add(record_uid)
            if record_uid not in params.record_cache:
                logging.warning(f'Record UID "{record_uid}" not found. Skipping.')
                continue
            if record_uid not in params.breach_watch_records:
                logging.warning(f'Record UID "{record_uid}": BreachWatch information not found')
                continue

        if len(record_uids) == 0:
            return

        bw_requests = []
        for record, password in params.breach_watch.get_records_by_status(params, ['WEAK', 'BREACHED']):
            if record.record_uid not in record_uids:
                continue
            record_uids.remove(record.record_uid)
            bwrq = breachwatch_pb2.BreachWatchRecordRequest()
            bwrq.recordUid = utils.base64_url_decode(record.record_uid)
            bwrq.breachWatchInfoType = breachwatch_pb2.RECORD
            bwrq.updateUserWhoScanned = False

            bw_password = client_pb2.BWPassword()
            bw_password.value = password.get('value')
            bw_password.resolved = utils.current_milli_time()
            bw_password.status = client_pb2.IGNORE
            euid = password.get('euid')
            if euid:
                bw_password.euid = base64.b64decode(euid)
            bw_data = client_pb2.BreachWatchData()
            bw_data.passwords.append(bw_password)
            data = bw_data.SerializeToString()
            try:
                record_key = params.record_cache[record.record_uid]['record_key_unencrypted']
                bwrq.encryptedData = crypto.encrypt_aes_v2(data, record_key)
            except:
                logging.warning(f'Record UID "{record.record_uid}" encryption error. Skipping.')
                continue
            bw_requests.append(bwrq)

        for record_uid in record_uids:
            logging.warning(f'Record UID "{record_uid}" cannot ignore. Skipping.')

        if bw_requests:
            params.sync_data = True
            if params.breach_watch.send_audit_events:
                params.queue_audit_event('bw_record_ignored')

            while bw_requests:
                chunk = bw_requests[0:999]
                bw_requests = bw_requests[999:]
                rq = breachwatch_pb2.BreachWatchUpdateRequest()
                rq.breachWatchRecordRequest.extend(chunk)
                rs = api.communicate_rest(params, rq, 'breachwatch/update_record_data',
                                          rs_type=breachwatch_pb2.BreachWatchUpdateResponse)
                for status in rs.breachWatchRecordStatus:
                    logging.info(f'{utils.base64_url_encode(status.recordUid)}: {status.status} {status.reason}')


class BreachWatchReportCommand(EnterpriseCommand):
    def get_parser(self):
        return breachwatch_report_parser

    def execute(self, params, **kwargs):
        BreachWatch.validate_reporting('breachwatch report', params)
        cmd = SecurityAuditReportCommand()
        return cmd.execute(params, **{'breachwatch': True, 'save': True, **kwargs})



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2025 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import queue
import threading
import time
import uuid
from datetime import datetime, timedelta
from enum import Enum
from typing import Dict, Any, Optional, Tuple
from dataclasses import dataclass, asdict

from ..util.command_util import CommandExecutor
from ..decorators.logging import logger, debug_decorator


# Queue configuration constants
DEFAULT_QUEUE_MAX_SIZE = 100
DEFAULT_REQUEST_TIMEOUT = 300  # 5 minutes in seconds
DEFAULT_RESULT_RETENTION = 3600  # 1 hour in seconds


class RequestStatus(Enum):
    """Request status enumeration."""
    QUEUED = "queued"
    PROCESSING = "processing" 
    COMPLETED = "completed"
    FAILED = "failed"
    EXPIRED = "expired"


@dataclass
class QueuedRequest:
    """Represents a queued command request."""
    request_id: str
    command: str
    status: RequestStatus
    created_at: datetime
    started_at: Optional[datetime] = None
    completed_at: Optional[datetime] = None
    result: Optional[Any] = None
    error_message: Optional[str] = None
    temp_files: list = None  # List of temporary file paths to clean up
    
    def to_dict(self) -> Dict[str, Any]:
        """Convert request to dictionary for JSON serialization."""
        data = asdict(self)
        # Convert datetime objects to ISO strings and handle bytes objects
        for key, value in data.items():
            if isinstance(value, datetime):
                data[key] = value.isoformat()
            elif isinstance(value, RequestStatus):
                data[key] = value.value
            elif isinstance(value, bytes):
                # Handle encrypted results - don't expose the raw bytes
                data[key] = {"encrypted": True, "type": "bytes"}
        return data


class RequestQueueManager:
    """Manages the request queue for sequential command processing."""
    
    _instance = None
    _lock = threading.Lock()
    
    def __new__(cls):
        """Singleton pattern implementation."""
        if cls._instance is None:
            with cls._lock:
                if cls._instance is None:
                    cls._instance = super().__new__(cls)
        return cls._instance
    
    def __init__(self):
        """Initialize the queue manager."""
        if hasattr(self, '_initialized'):
            return
            
        self._initialized = True
        # Initialize with internal constants
        self.request_queue = queue.Queue(maxsize=DEFAULT_QUEUE_MAX_SIZE)
        self.active_requests: Dict[str, QueuedRequest] = {}
        self.completed_requests: Dict[str, QueuedRequest] = {}
        self.worker_thread = None
        self.is_running = False
        self.current_request_id = None
        self.request_timeout = DEFAULT_REQUEST_TIMEOUT
        self.result_retention = DEFAULT_RESULT_RETENTION
        self.data_lock = threading.Lock()  # Lock for shared data structures
        
        logger.debug("RequestQueueManager initialized")
    
    @debug_decorator
    def start(self):
        """Start the queue processing worker thread."""
        if self.worker_thread and self.worker_thread.is_alive():
            logger.warning("Queue worker already running")
            return
        
        self.is_running = True
        self.worker_thread = threading.Thread(target=self._process_queue, daemon=True)
        self.worker_thread.start()
        logger.debug("Request queue worker started")
    
    @debug_decorator
    def stop(self):
        """Stop the queue processing worker thread."""
        self.is_running = False
        if self.worker_thread:
            self.worker_thread.join(timeout=5)
        logger.info("Request queue worker stopped")
    
    @debug_decorator
    def submit_request(self, command: str, temp_files: list = None) -> str:
        """Submit a new command request to the queue.
        
        Args:
            command: The command string to execute
            temp_files: List of temporary file paths to clean up after execution
            
        Returns:
            str: Unique request ID
            
        Raises:
            queue.Full: If the queue is at capacity
        """
        request_id = str(uuid.uuid4())
        request = QueuedRequest(
            request_id=request_id,
            command=command,
            status=RequestStatus.QUEUED,
            created_at=datetime.now(),
            temp_files=temp_files or []
        )
        
        try:
            self.request_queue.put(request, block=False)
            with self.data_lock:
                self.active_requests[request_id] = request
            logger.info(f"Request {request_id} queued: {command}")
            return request_id
        except queue.Full:
            logger.error("Error: Request queue is full")
            raise
    
    @debug_decorator
    def get_request_status(self, request_id: str) -> Optional[Dict[str, Any]]:
        """Get the status of a specific request.
        
        Args:
            request_id: The unique request identifier
            
        Returns:
            Dict containing request status and metadata, or None if not found
        """
        with self.data_lock:
            # Check active requests
            if request_id in self.active_requests:
                return self.active_requests[request_id].to_dict()
                
            # Check completed requests
            if request_id in self.completed_requests:
                return self.completed_requests[request_id].to_dict()
                
            return None
    
    @debug_decorator
    def get_request_result(self, request_id: str) -> Optional[Tuple[Any, int]]:
        """Get the result of a completed request.
        
        Args:
            request_id: The unique request identifier
            
        Returns:
            Tuple of (result, status_code) or None if not found/not completed
        """
        with self.data_lock:
            if request_id in self.completed_requests:
                request = self.completed_requests[request_id]
                if request.status == RequestStatus.COMPLETED:
                    return request.result, 200
                elif request.status == RequestStatus.FAILED:
                    return {"error": request.error_message}, 500
            return None
    
    @debug_decorator
    def get_queue_status(self) -> Dict[str, Any]:
        """Get overall queue status information.
        
        Returns:
            Dict containing queue statistics
        """
        with self.data_lock:
            return {
                "queue_size": self.request_queue.qsize(),
                "active_requests": len(self.active_requests),
                "completed_requests": len(self.completed_requests),
                "currently_processing": self.current_request_id,
                "worker_running": self.is_running and self.worker_thread and self.worker_thread.is_alive()
            }
    
    def _process_queue(self):
        """Main worker thread loop for processing queued requests."""
        logger.debug("Queue worker thread started")
        
        while self.is_running:
            try:
                # Get next request from queue (blocking with timeout)
                request = self.request_queue.get(timeout=1.0)
                self._process_request(request)
                self.request_queue.task_done()
                
            except queue.Empty:
                # Check for expired requests during idle time
                self._cleanup_expired_requests()
                continue
            except Exception as e:
                logger.error(f"Unexpected error in queue worker: {e}")
                time.sleep(1)
        
        logger.info("Queue worker thread stopped")
    
    def _process_request(self, request: QueuedRequest):
        """Process a single request.
        
        Args:
            request: The request to process
        """
        from ..util.request_validation import RequestValidator
        
        self.current_request_id = request.request_id
        request.status = RequestStatus.PROCESSING
        request.started_at = datetime.now()
        
        logger.info(f"Processing request {request.request_id}: {request.command}")
        
        try:
            # Execute the command using existing CommandExecutor
            result, status_code = CommandExecutor.execute(request.command)
            
            # Mark as completed
            request.status = RequestStatus.COMPLETED
            request.completed_at = datetime.now()
            request.result = result
            
            logger.info(f"Request {request.request_id} completed successfully")
            
        except Exception as e:
            # Mark as failed
            request.status = RequestStatus.FAILED
            request.completed_at = datetime.now()
            request.error_message = str(e)
            
            logger.error(f"Request {request.request_id} failed: {e}")
        
        finally:
            # Clean up temporary files
            if request.temp_files:
                RequestValidator.cleanup_temp_files(request.temp_files)
            
            # Move from active to completed
            with self.data_lock:
                if request.request_id in self.active_requests:
                    del self.active_requests[request.request_id]
                self.completed_requests[request.request_id] = request
                self.current_request_id = None
    
    def _cleanup_expired_requests(self):
        """Clean up expired and old completed requests."""
        now = datetime.now()
        
        with self.data_lock:
            # Find expired active requests
            expired_ids = []
            for request_id, request in self.active_requests.items():
                if request.status == RequestStatus.QUEUED:
                    age = (now - request.created_at).total_seconds()
                    if age > self.request_timeout:
                        request.status = RequestStatus.EXPIRED
                        expired_ids.append(request_id)
            
            # Move expired requests to completed
            for request_id in expired_ids:
                request = self.active_requests.pop(request_id)
                self.completed_requests[request_id] = request
                logger.warning(f"Request {request_id} expired after {self.request_timeout}s")
            
            # Clean up old completed requests
            cutoff_time = now - timedelta(seconds=self.result_retention)
            old_ids = []
            for request_id, request in self.completed_requests.items():
                if request.completed_at and request.completed_at < cutoff_time:
                    old_ids.append(request_id)
            
            for request_id in old_ids:
                del self.completed_requests[request_id]
                logger.debug(f"Cleaned up old request {request_id}")

queue_manager = RequestQueueManager()


#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Contact: ops@keepersecurity.com
#

import abc
import collections.abc
import datetime
import json
import logging
from typing import Optional, List, Tuple, Iterable, Type, Union, Dict, Any

import itertools

from .params import KeeperParams
from . import record_types, constants


def sanitize_str_field_value(value):    # type: (Any) -> str
    if not isinstance(value, str):
        value = str(value) if value else ''
    return value


def sanitize_int_field_value(value, *, default=0):    # type: (Any, *Any, Optional[Any]) -> int
    if not isinstance(value, int):
        try:
            value = int(value)
        except:
            if default is not None:
                if not isinstance(default, int):
                    default = 0
            value = default
    return value


def sanitize_bool_field_value(value):    # type: (Any) -> bool
    if not isinstance(value, bool):
        if isinstance(value, int):
            value = value != 0
        else:
            value = False
    return value


class KeeperRecord(abc.ABC):
    def __init__(self):
        self.record_uid = ''
        self.title = ''
        self.client_time_modified = 0
        self.record_key = b''
        self.revision = 0
        self.shared = False

    @abc.abstractmethod
    def get_version(self):  # type: () -> int
        pass

    @property
    def version(self):
        return self.get_version()

    @abc.abstractmethod
    def get_record_type(self):  # type: () -> str
        pass

    @property
    def record_type(self):
        return self.get_record_type()

    @abc.abstractmethod
    def load_record_data(self, data, extra=None):   # type: (dict, Optional[dict]) -> None
        pass

    @staticmethod
    def create(params, record_type):  # type: (KeeperParams, str) -> Optional['KeeperRecord']
        if not record_type:
            record_type = 'login'
        if record_type in {'legacy', 'general'}:
            record = PasswordRecord()
        elif record_type == 'file':
            record = FileRecord()
        else:
            record = TypedRecord()
            meta_data = None
            if params.record_type_cache:
                for rts in params.record_type_cache.values():
                    try:
                        rto = json.loads(rts)
                        if '$id' in rto and rto['$id'].lower() == record_type.lower():
                            meta_data = rto
                            break
                    except:
                        pass
            if meta_data and 'fields' in meta_data:
                for field in meta_data['fields']:
                    typed_field = TypedField()
                    typed_field.type = field.get('$ref', 'text')
                    typed_field.label = field.get('label', None)
                    record.fields.append(typed_field)
        return record

    @staticmethod
    def load(params, rec):
        # type: (KeeperParams, Union[str, Dict[str, Any]]) -> Optional['KeeperRecord']
        if isinstance(rec, str):
            if rec not in params.record_cache:
                return
            record = params.record_cache[rec]
        elif isinstance(rec, dict):
            record = rec
        else:
            return

        if 'data_unencrypted' not in record:
            return
        version = record.get('version', 0)

        if version == 2:
            keeper_record = PasswordRecord()
        elif version == 3:
            keeper_record = TypedRecord(version=3)
        elif version == 4:
            keeper_record = FileRecord()
            keeper_record.storage_size = record.get('file_size')
        elif version == 5:
            keeper_record = ApplicationRecord()
        elif version == 6:
            keeper_record = TypedRecord(version=6)
        else:
            return
        keeper_record.record_uid = record['record_uid']
        keeper_record.revision = record.get('revision', 0)
        keeper_record.record_key = record['record_key_unencrypted']
        keeper_record.client_time_modified = record.get('client_modified_time', 0)
        keeper_record.shared = record.get('shared', False)

        try:
            data = json.loads(record['data_unencrypted'])
        except:
            logging.warning('Record \"%s\": Corrupted record data', keeper_record.record_uid)
            return

        extra_str = record['extra_unencrypted'] if 'extra_unencrypted' in record else None
        extra = None     # type: Optional[Dict]
        if extra_str:
            try:
                extra = json.loads(extra_str)
            except:
                logging.debug('Record \"%s\": Corrupted record extra', keeper_record.record_uid)

        keeper_record.load_record_data(data, extra)

        return keeper_record

    def enumerate_fields(self):    # type: () -> Iterable[Tuple[str, Union[None, str, List[str]]]]
        yield '(title)', self.title

    @staticmethod
    def size_to_str(size):    # type: (Union[int, float]) -> str
        if isinstance(size, (int, float)):
            scale = 'Bytes'
            if size > 0:
                if size > 1000:
                    size = size / 1024
                    scale = 'Kb'
                if size > 1000:
                    size = size / 1024
                    scale = 'Mb'
                if size > 1000:
                    size = size / 1024
                    scale = 'Gb'
            return f'{size:.2f}'.rstrip('0').rstrip('.') + f' {scale}'
        elif isinstance(size, str):
            return size
        elif size:
            return str(size)
        else:
            return ''


class CustomField(object):
    def __init__(self, custom_field=None):  # type: (Optional[dict]) -> None
        if custom_field is None:
            custom_field = {}

        self.name = sanitize_str_field_value(custom_field.get('name')).strip()
        self.type = sanitize_str_field_value(custom_field.get('type')).strip().lower()
        self.value = sanitize_str_field_value(custom_field.get('value')).strip()

    @classmethod
    def new_field(cls, name, value):
        cf = CustomField()
        cf.type = 'text'
        cf.name = sanitize_str_field_value(name)
        cf.value = sanitize_str_field_value(value)
        return cf


class AttachmentFileThumb:
    def __init__(self, thumb_field=None):      # type: (Optional[dict]) -> None
        thumb_field = thumb_field or {}
        self.id = sanitize_str_field_value(thumb_field.get('id'))
        self.type = sanitize_str_field_value(thumb_field.get('type'))
        self.size = sanitize_int_field_value(thumb_field.get('size'))


class AttachmentFile(object):
    def __init__(self, file_field=None):  # type: (Optional[dict]) -> None
        file_field = file_field or {}
        self.id = sanitize_str_field_value(file_field.get('id'))
        self.key = sanitize_str_field_value(file_field.get('key'))
        self.name = sanitize_str_field_value(file_field.get('name'))
        self.title = sanitize_str_field_value(file_field.get('title'))
        self.mime_type = sanitize_str_field_value(file_field.get('type'))
        self.size = sanitize_int_field_value(file_field.get('size'))
        self.last_modified = sanitize_int_field_value(file_field.get('lastModified'), default=None)
        self.thumbnails = []         # type: List[AttachmentFileThumb]
        if file_field and 'thumbnails' in file_field:
            thumbs = file_field.get('thumbnails')
            if isinstance(thumbs, list):
                for thumb in thumbs:
                    self.thumbnails.append(AttachmentFileThumb(thumb))


class ExtraField(object):
    def __init__(self, extra_field=None):  # type: (Optional[dict]) -> None
        if extra_field is None:
            extra_field = {}
        self.id = sanitize_str_field_value(extra_field.get('id'))
        self.field_type = sanitize_str_field_value(extra_field.get('field_type'))
        self.field_title = sanitize_str_field_value(extra_field.get('field_title'))
        self.data = sanitize_str_field_value(extra_field.get('data'))


class PasswordRecord(KeeperRecord):
    def __init__(self):
        super(PasswordRecord, self).__init__()
        self.login = ''
        self.password = ''
        self.link = ''
        self.notes = ''
        self.custom = []         # type: List[CustomField]
        self.attachments = None  # type: Optional[List[AttachmentFile]]
        self.totp = ''

    def get_version(self):  # type: () -> int
        return 2

    def get_record_type(self):
        return ''

    def load_record_data(self, data, extra=None):
        self.title = sanitize_str_field_value(data.get('title')).strip()
        self.login = sanitize_str_field_value(data.get('secret1')).strip()
        self.password = sanitize_str_field_value(data.get('secret2'))
        self.link = sanitize_str_field_value(data.get('link'))
        self.notes = sanitize_str_field_value(data.get('notes'))
        custom = data.get('custom')
        if isinstance(custom, list):
            self.custom.extend((CustomField(x) for x in custom if isinstance(x, dict) and 'name' in x))
        if isinstance(extra, dict):
            if 'files' in extra:
                self.attachments = [AttachmentFile(x) for x in extra['files']]

            if 'fields' in extra and isinstance(extra['fields'], list):
                self.totp = next((x.get('data', '') for x in extra['fields'] if x.get('field_type') == 'totp'), '')

    def enumerate_fields(self):
        # type: () -> Iterable[Tuple[str, Union[None, str, List[str]]]]
        for pair in super(PasswordRecord, self).enumerate_fields():
            yield pair
        yield '(login)', self.login
        yield '(password)', self.password
        yield '(url)', self.link
        yield '(notes)', self.notes
        if self.totp:
            yield '(oneTimeCode)', self.totp
        for cf in self.custom:
            yield cf.name, cf.value
        if self.attachments:
            for atta in self.attachments:
                yield atta.title or atta.name, f'File ID: {atta.id}; Size: {KeeperRecord.size_to_str(atta.size)}'

    def get_custom_value(self, name):   # type: (str) -> Optional[str]
        field = next((x for x in self.custom if x.name == name), None)
        if field:
            return field.value

    def set_custom_value(self, name, value):   # type: (str, Optional[str]) -> None
        field = next((x for x in self.custom if x.name == name), None)
        value = sanitize_str_field_value(value)
        if value:
            if field:
                field.value = value
            else:
                self.custom.append(CustomField.new_field(name, value))
        else:
            if field:
                self.custom.remove(field)


class TypedField(object):
    def __init__(self, typed_field=None):
        if typed_field is None:
            typed_field = {}
        self.type = sanitize_str_field_value(typed_field.get('type')).strip()
        self.label = sanitize_str_field_value(typed_field.get('label')).strip()
        value = typed_field.get('value')
        if not isinstance(value, list):
            if isinstance(value, (str, int, bool)):
                value = [value]
            elif isinstance(value, collections.abc.Iterable):
                value = [x for x in value]
            else:
                value = []
        self.value = value
        self.required = sanitize_bool_field_value(typed_field.get('required'))

    @classmethod
    def new_field(cls, field_type, field_value, field_label=None):
        # type: (str, Any, Optional[str]) -> 'TypedField'
        f_type = sanitize_str_field_value(field_type) or 'text'
        # TODO check field value
        if not isinstance(field_value, list):
            if field_value:
                field_value = [field_value]
            else:
                field_value = []
        tf = TypedField()
        tf.type = f_type
        tf.value = field_value
        tf.label = sanitize_str_field_value(field_label)
        return tf

    def get_default_value(self, value_type=None):  # type: (Optional[Type]) -> any
        value = None
        if isinstance(self.value, list):
            if len(self.value) > 0:
                value = self.value[0]
        else:
            value = self.value
        if isinstance(value_type, type):
            if not isinstance(value, value_type):
                return
        return value

    def get_field_name(self):
        return f'({self.type}).{self.label}' if self.type and self.label else \
               f'({self.type})' if self.type else \
               f'{self.label}'

    @staticmethod
    def export_host_field(value):    # type: (dict) -> Optional[str]
        if isinstance(value, dict):
            host = value.get('hostName') or ''
            port = value.get('port') or ''
            if host or port:
                if port:
                    host += ':' + str(port)
            return host

    @staticmethod
    def export_phone_field(value):   # type: (dict) -> Optional[str]
        if isinstance(value, dict):
            phone = value.get('type') or ''
            if phone:
                phone += ':'
            region = value.get('region') or ''
            if region:
                if len(region) == 2 and region.isalpha():
                    pass
                elif region.isnumeric():
                    region = '+' + region
                else:
                    region = ''
                if region:
                    phone += '  ' + region
            number = (value.get('number') or '').replace(' ', '-')
            if number:
                phone += ' ' + number
            ext = (value.get('ext') or '').replace(' ', '-')
            if ext:
                phone += ' ' + ext
            return phone

    @staticmethod
    def export_name_field(value):   # type: (dict) -> Optional[str]
        if isinstance(value, dict):
            first_name = value.get('first') or ''
            middle_name = value.get('middle') or ''
            name = value.get('last') or ''
            if first_name or middle_name or name:
                name = f'{name}, {first_name}'
                if middle_name:
                    name += ' ' + middle_name
            return name

    @staticmethod
    def export_address_field(value):   # type: (dict) -> Optional[str]
        if isinstance(value, dict):
            address = value.get('street1', '').replace(',', '.')
            street2 = value.get('street2', '').replace(',', '.')
            if street2:
                address += ' ' + street2
            city = value.get('city', '').replace(',', '.')
            if city:
                address += ', ' + city
                state = value.get('state', '').replace(',', '.')
                zip_code = value.get('zip', '').replace(',', '.')
                if state or zip_code:
                    address += ', ' + state + ' ' + zip_code
                    country = value.get('country', '').replace(',', '.')
                    if country:
                        address += ', ' + country
            return address

    @staticmethod
    def export_q_and_a_field(value):   # type: (dict) -> Optional[str]
        if isinstance(value, dict):
            q = value.get('question', '').replace('?', '')
            a = value.get('answer', '')
            return f'{q}? {a}'.strip()

    @staticmethod
    def export_card_field(value):   # type: (dict) -> Optional[str]
        if isinstance(value, dict):
            comps = []
            number = value.get('cardNumber')
            if number:
                comps.append(number)
            expiration = value.get('cardExpirationDate')
            if expiration:
                comps.append(expiration)
            cvv = value.get('cardSecurityCode')
            if cvv:
                comps.append(cvv)
            if comps:
                return ' '.join(comps)

    @staticmethod
    def export_account_field(value):   # type: (dict) -> Optional[str]
        if isinstance(value, dict):
            account_type = value.get('accountType', '').replace(' ', '')
            routing = value.get('routingNumber', '').replace(' ', '')
            account_number = value.get('accountNumber', '').replace(' ', '')
            if routing or account_number:
                comps = []
                if account_type:
                    comps.append(account_type)
                if routing:
                    comps.append(routing)
                if account_number:
                    comps.append(account_number)
                if comps:
                    return ' '.join(comps)

    @staticmethod
    def export_ssh_key_field(value):   # type: (dict) -> Optional[str]
        if isinstance(value, dict):
            return value.get('privateKey', '')

    @staticmethod
    def export_schedule_field(value):   # type: (dict) -> Optional[str]
        if isinstance(value, dict):
            schedule_type = value.get('type')
            if schedule_type == 'RUN_ONCE':
                return ''

            if schedule_type == 'CRON':
                cron = value.get('cron')
                if isinstance(cron, str):
                    comps = [x for x in cron.split(' ') if x]
                    if len(comps) >= 6:
                        comps = comps[1:6]
                        return ' '.join(comps)
                return ''

            hour = '0'
            minute = '0'
            day = '*'
            month = '*'
            week_day = '*'
            utc_time = value.get('time')
            if not utc_time:
                utc_time = value.get('utcTime')
            if utc_time:
                comps = utc_time.split(':')
                if len(comps) >= 2:
                    if comps[0].isnumeric():
                        h = int(comps[0])
                        if 0 <= h <= 23:
                            hour = str(h)
                    if comps[1].isnumeric():
                        m = int(comps[1])
                        if 0 <= m <= 59:
                            minute = str(m)

            if schedule_type == 'DAILY':
                interval = value.get('intervalCount') or 0
                if interval > 1:
                    if interval > 28:
                        interval = 28
                    day = f'*/{interval}'
            elif schedule_type == 'WEEKLY':
                week_day = constants.get_cron_week_day(value.get('weekday')) or 1
            elif schedule_type == 'MONTHLY_BY_DAY':
                day = constants.get_cron_month_day(value.get('monthDay')) or 1
            elif schedule_type == 'MONTHLY_BY_WEEKDAY':
                wd = constants.get_cron_week_day(value.get('weekday')) or 1
                occ = constants.get_cron_occurrence(value.get('occurrence'))
                if occ == 'FIRST':
                    occ = 1
                elif occ == 'SECOND':
                    occ = 2
                elif occ == 'THIRD':
                    occ = 3
                elif occ == 'FOURTH':
                    occ = 4
                else:
                    occ = 1
                week_day = f'{wd}#{occ}'
            elif schedule_type == 'YEARLY':
                month = constants.get_cron_month(value.get('month')) or 1
                day = constants.get_cron_month_day(value.get('monthDay')) or 1
            else:
                return ''

            return f'{minute} {hour} {day} {month} {week_day}'

    @staticmethod
    def import_host_field(value):    # type: (str) -> Optional[dict]
        if isinstance(value, str):
            host, _, port = value.partition(':')
            return {
                'hostName': host.strip(),
                'port': port.strip()
            }

    @staticmethod
    def import_phone_field(value):   # type: (str) -> Optional[dict]
        if isinstance(value, str):
            region = ''
            number = ''
            ext = ''
            phone_type, _, rest = value.partition(':')
            if not rest:
                rest = phone_type
                phone_type = ''
            comps = rest.strip().split(' ')
            for comp in comps:
                comp = comp.strip()
                if comp.isalpha():
                    if len(comp) == 2:
                        if not region:
                            region = comp
                    elif not phone_type:
                        phone_type = comp
                elif len(comp) >= 6:
                    if not number:
                        number = comp
                elif not ext:
                    ext = comp
            result = {
                'type': '',
                'region': '',
                'number': number.strip(),
                'ext': ext.strip()
            }
            phone_type = phone_type.strip()
            region = region.strip()
            if phone_type:
                result['type'] = phone_type
            if region:
                result['region'] = region
            return result

    @staticmethod
    def import_name_field(value):   # type: (str) -> Optional[dict]
        if isinstance(value, str):
            first = ''
            middle = ''
            last = ''
            comma_pos = value.find(',')
            if comma_pos >= 0:
                last = value[:comma_pos]
                rest = value[comma_pos+1:]
            else:
                space_pos = value.rfind(' ')
                if space_pos >= 0:
                    last = value[space_pos+1:]
                    rest = value[:space_pos]
                else:
                    last = value
                    rest = ''
            rest = rest.strip()
            if rest:
                space_pos = rest.rfind(' ')
                if space_pos >= 0:
                    middle = rest[space_pos+1:]
                    first = rest[:space_pos]
                else:
                    middle = ''
                    first = rest

            return {
                'first': first.strip(),
                'middle': middle.strip(),
                'last': last.strip(),
            }

    @staticmethod
    def import_address_field(value):   # type: (str) -> Optional[dict]
        if isinstance(value, str):
            comps = value.split(',')
            street1 = comps[0].strip() if len(comps) > 0 else ''
            city = comps[1].strip() if len(comps) > 1 else ''
            state, _, zip_code = comps[2].strip().partition(' ') if len(comps) > 2 else ('', '', '')
            if state and not zip_code:
                if state.isnumeric():
                    zip_code = state
                    state = ''
            country = comps[3].strip() if len(comps) > 3 else ''

            return {
                'street1': street1,
                'street2': '',
                'city': city,
                'state': state,
                'zip': zip_code,
                'country': country
            }

    @staticmethod
    def import_q_and_a_field(value):   # type: (str) -> Optional[dict]
        if isinstance(value, str):
            q, sign, a = value.partition('?')
            return {
                'question': q.strip() + '?',
                'answer': a.strip(),
            }

    @staticmethod
    def import_card_field(value):   # type: (str) -> Optional[dict]
        if isinstance(value, str):
            comps = value.split(' ')
            number = ''
            expiration = ''
            cvv = ''
            for comp in comps:
                comp = comp.strip()
                if comp:
                    if len(comp) > 10:
                        number = comp
                    elif comp.find('/') >= 0:
                        expiration = comp
                    elif len(comp) <= 6:
                        cvv = comp
            if number or expiration or cvv:
                return {
                    'cardNumber': number,
                    'cardExpirationDate': expiration,
                    'cardSecurityCode':  cvv,
                }

    @staticmethod
    def import_account_field(value):   # type: (str) -> Optional[dict]
        if isinstance(value, str):
            account_type = ''
            routing = ''
            account_number = ''
            comps = value.split()
            for comp in comps:
                comp = comp.strip()
                if comp.isnumeric():
                    if not routing:
                        routing = comp
                    elif not account_number:
                        account_number = comp
                else:
                    if not account_type:
                        account_type = comp
            if routing or account_number:
                return {
                    'accountType': account_type,
                    'routingNumber': routing,
                    'accountNumber': account_number
                }

    @staticmethod
    def import_ssh_key_field(value):   # type: (str) -> Optional[dict]
        if isinstance(value, str):
            return {
                'privateKey': value.replace('\\n', '\n'),
                'publicKey': ''
            }

    @staticmethod
    def import_schedule_field(value):   # type: (str) -> Optional[dict]
        if isinstance(value, str) and len(value) > 0:
            comps = value.split(' ')
            if len(comps) >= 3:
                schedule = None
                if len(comps) < 4:
                    comps.insert(0, '0')
                if len(comps) < 5:
                    comps.insert(0, '0')
                minute = int(comps[0]) if comps[0].isnumeric() else 0
                if minute < 0 or minute > 59:
                    minute = 0
                hour = 0
                if comps[1].isnumeric():
                    hour = int(comps[1])
                else:
                    hrs = comps[1].replace('-', ',').split(',')
                    if len(hrs) > 0 and hrs[0].isnumeric():
                        hour = int(hrs[0])
                if hour < 0 or hour > 23:
                    hour = 0
                time = f'{hour:02}:{minute:02}:00'
                intervalCount = 1

                if comps[3] in ('*', '?') and comps[4] in ('*', '?'):  # daily
                    if comps[2].isnumeric():
                        schedule = {
                            'type': 'MONTHLY_BY_DAY',
                            'monthDay': int(comps[2])
                        }
                    elif comps[2].startswith('*'):
                        schedule = {
                            'type': 'DAILY',
                        }
                        if comps[2].startswith('*/'):
                            intr = comps[2][2:]
                            if intr.isnumeric():
                                schedule['occurrences'] = int(intr)
                elif comps[4] not in ('*', '?'):  # day of week
                    if comps[4].isnumeric():
                        wd = int(comps[4])
                        if wd < 0 or wd > len(constants.week_days):
                            wd = 1
                        schedule = {
                            'type': 'WEEKLY',
                            'weekday': constants.week_days[wd]
                        }
                    elif comps[4].startswith('*/'):
                        schedule = {
                            'type': 'DAILY',
                        }
                        intr = comps[4][2:]
                        if intr.isnumeric():
                            schedule['occurrences'] = int(intr)
                    else:
                        wd_comps = comps[4].replace('%', '#').split('#')
                        if len(wd_comps) == 2 and wd_comps[0].isnumeric() and wd_comps[1].isnumeric():
                            wd = int(wd_comps[0])
                            if wd < 0 or wd > len(constants.week_days):
                                wd = 1
                            occ = int(wd_comps[1]) - 1
                            if occ < 0 or occ >= len(constants.occurrences):
                                occ = 0
                            occurrence = constants.occurrences[occ]
                            schedule = {
                                'type': 'MONTHLY_BY_WEEKDAY',
                                'weekday': constants.week_days[wd],
                                'occurrence': occurrence,
                            }
                elif comps[3].isnumeric():  # day of month
                    mm = int(comps[3])
                    if mm > 0:
                        mm -= 1
                        if mm >= len(constants.months):
                            mm = len(constants.months) - 1
                    else:
                        mm = 0

                    dd = 1
                    if comps[2].isnumeric():
                        dd = int(comps[2])

                    schedule = {
                        'type': 'YEARLY',
                        'month': constants.months[mm],
                        'monthDay': dd,
                    }
                else:
                    schedule = {
                        'type': 'RUN_ONCE',
                    }
                    time = '2000-01-01T00:00:00'

                schedule['tz'] = 'Etc/UTC'
                schedule['time'] = time
                schedule['intervalCount'] = intervalCount
                return schedule

    @staticmethod
    def get_exported_value(field_type, field_value):
        # type: (str, Any) -> Iterable[str]
        if not field_value:
            return

        if isinstance(field_value, str):
            yield field_value
            return

        rf = record_types.RecordFields.get(field_type)
        ft = record_types.FieldTypes.get(rf.type) if rf else None
        if isinstance(field_value, int):
            if ft and ft.name == 'date':
                if field_value != 0:
                    dt = datetime.datetime.fromtimestamp(int(field_value // 1000)).date()
                    yield str(dt)
            else:
                yield str(field_value)
        elif isinstance(field_value, list):
            for elem in field_value:
                for ev in TypedField.get_exported_value(field_type, elem):
                    yield ev
        elif isinstance(field_value, dict):
            if ft:
                if ft.name == 'host':
                    yield TypedField.export_host_field(field_value)
                elif ft.name == 'phone':
                    yield TypedField.export_phone_field(field_value)
                elif ft.name == 'name':
                    yield TypedField.export_name_field(field_value)
                elif ft.name == 'address':
                    yield TypedField.export_address_field(field_value)
                elif ft.name == 'securityQuestion':
                    yield TypedField.export_q_and_a_field(field_value)
                elif ft.name == 'paymentCard':
                    yield TypedField.export_card_field(field_value)
                elif ft.name == 'bankAccount':
                    yield TypedField.export_account_field(field_value)
                elif ft.name == 'privateKey':
                    yield TypedField.export_ssh_key_field(field_value)
                elif ft.name == 'schedule':
                    yield TypedField.export_schedule_field(field_value)

    def get_external_value(self):   # type: () -> Iterable[str]
        for value in self.get_exported_value(self.type, self.value):
            if value:
                yield value


class TypedRecord(KeeperRecord):
    def __init__(self, version=3):
        super(TypedRecord, self).__init__()
        self._version = version
        self.type_name = ''
        self.notes = ''
        self.fields = []         # type: List[TypedField]
        self.custom = []         # type: List[TypedField]
        self.linked_keys = None  # type: Optional[Dict[str, bytes]]

    def get_version(self):  # type: () -> int
        return self._version

    def get_record_type(self):
        return self.type_name

    def get_typed_field(self, field_type, label=None):    # type: (str, Optional[str]) -> Optional[TypedField]
        return next((x for x in itertools.chain(self.fields, self.custom)
                     if field_type == x.type and (not label or (x.label and label.casefold() == x.label.casefold()))),
                    None)

    def load_record_data(self, data, extra=None):
        self.type_name = sanitize_str_field_value(data.get('type')).strip()
        self.title = sanitize_str_field_value(data.get('title')).strip()
        self.notes = sanitize_str_field_value(data.get('notes'))
        self.fields.extend((TypedField(x) for x in data.get('fields', [])))
        self.custom.extend((TypedField(x) for x in data.get('custom', [])))

    def enumerate_fields(self):
        # type: () -> Iterable[Tuple[str, Union[None, str, List[str]]]]
        for pair in super(TypedRecord, self).enumerate_fields():
            yield pair
        yield '(type)', self.record_type
        yield '(notes)', self.notes
        for field in itertools.chain(self.fields, self.custom):
            values = list(field.get_external_value())
            yield field.get_field_name(), '' if len(values) == 0 else values[0] if len(values) == 1 else values


class FileRecord(KeeperRecord):
    def __init__(self):
        super(FileRecord, self).__init__()
        self.name = ''
        self.size = None   # type: Optional[int]
        self.mime_type = ''
        self.last_modified = None   # type: Optional[int]
        self.storage_size = None   # type: Optional[int]

    def get_version(self):  # type: () -> int
        return 4

    def get_record_type(self):
        return 'file'

    def load_record_data(self, data, extra=None):
        self.title = sanitize_str_field_value(data.get('title')).strip()
        self.name = sanitize_str_field_value(data.get('name')).strip()
        self.size = sanitize_int_field_value(data.get('size'))
        self.mime_type = sanitize_str_field_value(data.get('type'))
        self.last_modified = sanitize_int_field_value(data.get('lastModified'), default=None)

    def enumerate_fields(self):  # type: () -> Iterable[Tuple[str, Union[None, str, List[str]]]]
        for pair in super(FileRecord, self).enumerate_fields():
            yield pair
        yield '(type)', self.get_record_type()
        yield '(name)', self.name
        if self.mime_type:
            yield '(mime-type)', self.mime_type
        yield '(size)', KeeperRecord.size_to_str(self.size)


class ApplicationRecord(KeeperRecord):
    def __init__(self):
        super(ApplicationRecord, self).__init__()

    def get_version(self):
        return 5

    def get_record_type(self):
        return 'app'

    def load_record_data(self, data, extra=None):
        self.title = sanitize_str_field_value(data.get('title')).strip()

    def enumerate_fields(self):  # type: () -> Iterable[Tuple[str, Union[None, str, List[str]]]]
        for pair in super(ApplicationRecord, self).enumerate_fields():
            yield pair
        yield '(type)', self.get_record_type()



import json
import getpass
import logging
import pyperclip
import re
import webbrowser
from typing import Optional, List

from . import login_steps
from .. import utils
from ..display import bcolors
from ..error import KeeperApiError


class ConsoleLoginUi(login_steps.LoginUi):
    def __init__(self):
        self._show_device_approval_help = True
        self._show_two_factor_help = True
        self._show_password_help = True
        self._show_sso_redirect_help = True
        self._show_sso_data_key_help = True
        self._failed_password_attempt = 0

    def on_device_approval(self, step):
        if self._show_device_approval_help:
            print("\nDevice Approval Required\n")

            print("Approve by selecting a method below:")
            print("\t\"" + bcolors.OKGREEN + "email_send" + bcolors.ENDC + "\" to send email")
            print("\t\"" + bcolors.OKGREEN + "email_code=<code>" + bcolors.ENDC + "\" to validate verification code sent via email")
            print("\t\"" + bcolors.OKGREEN + "keeper_push" + bcolors.ENDC + "\" to send Keeper Push notification")
            print("\t\"" + bcolors.OKGREEN + "2fa_send" + bcolors.ENDC + "\" to send 2FA code")
            print("\t\"" + bcolors.OKGREEN + "2fa_code=<code>" + bcolors.ENDC + "\" to validate a code provided by 2FA application")
            print("\t\"" + bcolors.OKGREEN + "<Enter>" + bcolors.ENDC + "\" to resume")

            self._show_device_approval_help = False
        else:
            print(bcolors.BOLD + "\nWaiting for device approval." + bcolors.ENDC)
            print("Check email, SMS message or push notification on the approved device.\n")

        try:
            selection = input('Type your selection or <Enter> to resume: ')

            if selection == "email_send" or selection == "es":
                step.send_push(login_steps.DeviceApprovalChannel.Email)
                print(bcolors.WARNING + "\nAn email with instructions has been sent to " + step.username + bcolors.WARNING + '\nPress <Enter> when approved.')

            elif selection.startswith("email_code="):
                code = selection.replace("email_code=", "")
                step.send_code(login_steps.DeviceApprovalChannel.Email, code)
                print("Successfully verified email code.")

            elif selection == "2fa_send" or selection == "2fs":
                step.send_push(login_steps.DeviceApprovalChannel.TwoFactor)
                print(bcolors.WARNING + "\n2FA code was sent." + bcolors.ENDC)

            elif selection.startswith("2fa_code="):
                code = selection.replace("2fa_code=", "")
                step.send_code(login_steps.DeviceApprovalChannel.TwoFactor, code)
                print("Successfully verified 2FA code.")

            elif selection == "keeper_push" or selection == "kp":
                step.send_push(login_steps.DeviceApprovalChannel.KeeperPush)
                logging.info('Successfully made a push notification to the approved device.\nPress <Enter> when approved.')

            elif selection == "":
                step.resume()
        except KeyboardInterrupt:
            step.cancel()
        except KeeperApiError as kae:
            print()
            print(bcolors.WARNING + kae.message + bcolors.ENDC)
            pass

    @staticmethod
    def two_factor_channel_to_desc(channel):   # type: (login_steps.TwoFactorChannel) -> str
        if channel == login_steps.TwoFactorChannel.Authenticator:
            return 'TOTP (Google and Microsoft Authenticator)'
        if channel == login_steps.TwoFactorChannel.TextMessage:
            return 'Send SMS Code'
        if channel == login_steps.TwoFactorChannel.DuoSecurity:
            return 'DUO'
        if channel == login_steps.TwoFactorChannel.RSASecurID:
            return 'RSA SecurID'
        if channel == login_steps.TwoFactorChannel.SecurityKey:
            return 'WebAuthN (FIDO2 Security Key)'
        if channel == login_steps.TwoFactorChannel.KeeperDNA:
            return 'Keeper DNA (Watch)'
        if channel == login_steps.TwoFactorChannel.Backup:
            return 'Backup Codes'

    def on_two_factor(self, step):
        channels = step.get_channels()

        if self._show_two_factor_help:
            print("\nThis account requires 2FA Authentication\n")
            for i in range(len(channels)):
                channel = channels[i]
                print(f"{i+1:>3}. {ConsoleLoginUi.two_factor_channel_to_desc(channel.channel_type)} {channel.channel_name} {channel.phone}")
            print(f"{'q':>3}. Quit login attempt and return to Commander prompt")
            self._show_device_approval_help = False

        channel = None    # type: Optional[login_steps.TwoFactorChannelInfo]
        while channel is None:
            selection = input('Selection: ')
            if selection == 'q':
                raise KeyboardInterrupt()

            if selection.isnumeric():
                idx = int(selection)
                if 1 <= idx <= len(channels):
                    channel = channels[idx-1]
                    logging.debug(f"Selected {idx}. {ConsoleLoginUi.two_factor_channel_to_desc(channel.channel_type)}")
                else:
                    print("Invalid entry, additional factors of authentication shown may be configured if not currently enabled.")
            else:
                print("Invalid entry, additional factors of authentication shown may be configured if not currently enabled.")

        mfa_prompt = False

        if channel.channel_type == login_steps.TwoFactorChannel.Other:
            pass
        elif channel.channel_type == login_steps.TwoFactorChannel.TextMessage:
            mfa_prompt = True
            try:
                step.send_push(channel.channel_uid, login_steps.TwoFactorPushAction.TextMessage)
                print(bcolors.OKGREEN + "\nSuccessfully sent SMS.\n" + bcolors.ENDC)
            except KeeperApiError:
                print("Was unable to send SMS.")
        elif channel.channel_type == login_steps.TwoFactorChannel.SecurityKey:
            try:
                from ..yubikey.yubikey import yubikey_authenticate
                challenge = json.loads(channel.challenge)
                response = yubikey_authenticate(challenge)

                if response:
                    signature = {
                        "id": response.id,
                        "rawId": utils.base64_url_encode(response.raw_id),
                        "response": {
                            "authenticatorData": utils.base64_url_encode(response.response.authenticator_data),
                            "clientDataJSON": response.response.client_data.b64,
                            "signature": utils.base64_url_encode(response.response.signature),
                        },
                        "type": "public-key",
                        "clientExtensionResults": dict(response.client_extension_results) if response.client_extension_results else {}
                    }
                    step.duration = login_steps.TwoFactorDuration.EveryLogin
                    step.send_code(channel.channel_uid, json.dumps(signature))
                    print(bcolors.OKGREEN + "Verified Security Key." + bcolors.ENDC)

            except ImportError as e:
                from ..yubikey import display_fido2_warning
                display_fido2_warning()
                logging.warning(e)
            except KeeperApiError:
                print(bcolors.FAIL + "Unable to verify code generated by security key" + bcolors.ENDC)
            except Exception as e:
                logging.error(e)

        elif channel.channel_type in {login_steps.TwoFactorChannel.Authenticator,
                                      login_steps.TwoFactorChannel.DuoSecurity,
                                      login_steps.TwoFactorChannel.RSASecurID,
                                      login_steps.TwoFactorChannel.KeeperDNA,
                                      login_steps.TwoFactorChannel.Backup}:
            mfa_prompt = True
        else:
            raise NotImplementedError(f"Unhandled channel type {ConsoleLoginUi.two_factor_channel_to_desc(channel.channel_type)}")

        if mfa_prompt:
            config_expiration = step.get_max_duration()
            mfa_expiration = step.duration

            if mfa_expiration > config_expiration:
                mfa_expiration = config_expiration

            allowed_expirations = ['login']     # type: List[str]
            if channel.max_expiration >= login_steps.TwoFactorDuration.Every12Hours:
                allowed_expirations.append('12_hours')
            if channel.max_expiration >= login_steps.TwoFactorDuration.Every24Hours:
                allowed_expirations.append('24_hours')
            if channel.max_expiration >= login_steps.TwoFactorDuration.Every30Days:
                allowed_expirations.append('30_days')
            if channel.max_expiration >= login_steps.TwoFactorDuration.Forever:
                allowed_expirations.append('forever')

            otp_code = ''
            show_duration = True
            mfa_pattern = re.compile(r'2fa_duration\s*=\s*(.+)', re.IGNORECASE)
            while not otp_code:
                if show_duration:
                    show_duration = False
                    prompt_exp = '\n2FA Code Duration: {0}.\nTo change duration: 2fa_duration={1}'.format(
                        'Require Every Login' if mfa_expiration == login_steps.TwoFactorDuration.EveryLogin else
                        'Save on this Device Forever' if mfa_expiration == login_steps.TwoFactorDuration.Forever else
                        'Ask Every 12 hours' if mfa_expiration == login_steps.TwoFactorDuration.Every12Hours else
                        'Ask Every 24 hours' if mfa_expiration == login_steps.TwoFactorDuration.Every24Hours else
                        'Ask Every 30 days',
                        "|".join(allowed_expirations))
                    print(prompt_exp)

                try:
                    answer = input('\nEnter 2FA Code or Duration: ')
                except KeyboardInterrupt:
                    step.cancel()
                    return

                m_duration = re.match(mfa_pattern, answer)
                if m_duration:
                    answer = m_duration.group(1).strip().lower()
                    if answer not in allowed_expirations:
                        print(f'Invalid 2FA Duration: {answer}')
                        answer = ''

                if answer == 'login':
                    show_duration = True
                    mfa_expiration = login_steps.TwoFactorDuration.EveryLogin
                elif answer == '12_hours':
                    show_duration = True
                    mfa_expiration = login_steps.TwoFactorDuration.Every12Hours
                elif answer == '24_hours':
                    show_duration = True
                    mfa_expiration = login_steps.TwoFactorDuration.Every24Hours
                elif answer == '30_days':
                    show_duration = True
                    mfa_expiration = login_steps.TwoFactorDuration.Every30Days
                elif answer == 'forever':
                    show_duration = True
                    mfa_expiration = login_steps.TwoFactorDuration.Forever
                else:
                    otp_code = answer

            step.duration = mfa_expiration
            try:
                step.send_code(channel.channel_uid, otp_code)
                print(bcolors.OKGREEN + "Successfully verified 2FA Code." + bcolors.ENDC)
            except KeeperApiError:
                warning_msg = bcolors.WARNING + f"Unable to verify 2FA code. Regenerate the code and try again." + bcolors.ENDC
                print(warning_msg)

    def on_password(self, step):
        if self._show_password_help:
            print(f'Enter password for {step.username}')

        if self._failed_password_attempt > 0:
            print('Forgot password? Type "recover"<Enter>')

        password = getpass.getpass(prompt='Password: ', stream=None)
        if not password:
            step.cancel()
        elif password == 'recover':
            step.forgot_password()
        else:
            try:
                step.verify_password(password)
            except KeeperApiError as kae:
                print(kae.message)
            except KeyboardInterrupt:
                step.cancel()

    def on_sso_redirect(self, step):
        try:
            wb = webbrowser.get()
            wrappers = set('xdg-open|gvfs-open|gnome-open|x-www-browser|www-browser'.split('|'))
            browsers = set(webbrowser._browsers if hasattr(webbrowser, '_browsers') else {})
            standalones = browsers - wrappers
            if browsers and not standalones:     # show browser-launch option only if effectively supported
                wb = None
        except:
            wb = None

        sp_url = step.sso_login_url
        print(f'\nSSO Login URL:\n{sp_url}\n')
        if self._show_sso_redirect_help:
            print('Navigate to SSO Login URL with your browser and complete login.')
            print('Copy a returned SSO Token into clipboard.')
            print('Paste that token into Commander')
            print('NOTE: To copy SSO Token please click "Copy login token" button on "SSO Connect" page.')
            print('')
            print('  a. SSO User with a Master Password')
            print('  c. Copy SSO Login URL to clipboard')
            if wb:
                print('  o. Navigate to SSO Login URL with the default web browser')
            print('  p. Paste SSO Token from clipboard')
            print('  q. Quit SSO login attempt and return to Commander prompt')
            self._show_sso_redirect_help = False

        while True:
            try:
                token = input('Selection: ')
            except KeyboardInterrupt:
                step.cancel()
                return
            if token == 'q':
                step.cancel()
                return
            if token == 'a':
                step.login_with_password()
                return
            if token == 'c':
                token = None
                try:
                    pyperclip.copy(sp_url)
                    print('SSO Login URL is copied to clipboard.')
                except:
                    print('Failed to copy SSO Login URL to clipboard.')
            elif token == 'o':
                token = None
                if wb:
                    try:
                        wb.open_new_tab(sp_url)
                    except:
                        print('Failed to open web browser.')
            elif token == 'p':
                try:
                    token = pyperclip.paste()
                except:
                    token = ''
                    logging.info('Failed to paste from clipboard')
            else:
                if len(token) < 10:
                    print(f'Unsupported menu option: {token}')
                    token = None
            if token:
                step.set_sso_token(token)
                break

    def on_sso_data_key(self, step):
        if self._show_sso_data_key_help:
            print('\nApprove this device by selecting a method below:')
            print('  1. Keeper Push. Send a push notification to your device.')
            print('  2. Admin Approval. Request your admin to approve this device.')
            print('')
            print('  r. Resume SSO login after device is approved.')
            print('  q. Quit SSO login attempt and return to Commander prompt.')
            self._show_sso_data_key_help = False

        while True:
            try:
                answer = input('Selection: ')
            except KeyboardInterrupt:
                answer = 'q'

            if answer == 'q':
                step.cancel()
                break
            elif answer == 'r':
                step.resume()
                break
            elif answer == '1':
                step.request_data_key(login_steps.DataKeyShareChannel.KeeperPush)
            elif answer == '2':
                step.request_data_key(login_steps.DataKeyShareChannel.AdminApproval)
            else:
                print(f'Action \"{answer}\" is not supported.')



import ssocloud_pb2 as _ssocloud_pb2
import enterprise_pb2 as _enterprise_pb2
import version_pb2 as _version_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class SsoAuthenticationProtocolType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    UNKNOWN_PROTOCOL: _ClassVar[SsoAuthenticationProtocolType]
    SAML2: _ClassVar[SsoAuthenticationProtocolType]

class CertificateFormat(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    UNKNOWN_FORMAT: _ClassVar[CertificateFormat]
    PKCS12: _ClassVar[CertificateFormat]
    JKS: _ClassVar[CertificateFormat]

class SkillType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    UNKNOWN_SKILL_TYPE: _ClassVar[SkillType]
    DEVICE_APPROVAL: _ClassVar[SkillType]
    TEAM_APPROVAL: _ClassVar[SkillType]
    TEAM_FOR_USER_APPROVAL: _ClassVar[SkillType]

class AutomatorState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    UNKNOWN_STATE: _ClassVar[AutomatorState]
    RUNNING: _ClassVar[AutomatorState]
    ERROR: _ClassVar[AutomatorState]
    NEEDS_INITIALIZATION: _ClassVar[AutomatorState]
    NEEDS_CRYPTO_STEP_1: _ClassVar[AutomatorState]
    NEEDS_CRYPTO_STEP_2: _ClassVar[AutomatorState]
UNKNOWN_PROTOCOL: SsoAuthenticationProtocolType
SAML2: SsoAuthenticationProtocolType
UNKNOWN_FORMAT: CertificateFormat
PKCS12: CertificateFormat
JKS: CertificateFormat
UNKNOWN_SKILL_TYPE: SkillType
DEVICE_APPROVAL: SkillType
TEAM_APPROVAL: SkillType
TEAM_FOR_USER_APPROVAL: SkillType
UNKNOWN_STATE: AutomatorState
RUNNING: AutomatorState
ERROR: AutomatorState
NEEDS_INITIALIZATION: AutomatorState
NEEDS_CRYPTO_STEP_1: AutomatorState
NEEDS_CRYPTO_STEP_2: AutomatorState

class AutomatorSettingValue(_message.Message):
    __slots__ = ["settingId", "settingTypeId", "settingTag", "settingName", "settingValue", "dataType", "lastModified", "fromFile", "encrypted", "encoded", "editable", "translated", "userVisible", "required"]
    SETTINGID_FIELD_NUMBER: _ClassVar[int]
    SETTINGTYPEID_FIELD_NUMBER: _ClassVar[int]
    SETTINGTAG_FIELD_NUMBER: _ClassVar[int]
    SETTINGNAME_FIELD_NUMBER: _ClassVar[int]
    SETTINGVALUE_FIELD_NUMBER: _ClassVar[int]
    DATATYPE_FIELD_NUMBER: _ClassVar[int]
    LASTMODIFIED_FIELD_NUMBER: _ClassVar[int]
    FROMFILE_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTED_FIELD_NUMBER: _ClassVar[int]
    ENCODED_FIELD_NUMBER: _ClassVar[int]
    EDITABLE_FIELD_NUMBER: _ClassVar[int]
    TRANSLATED_FIELD_NUMBER: _ClassVar[int]
    USERVISIBLE_FIELD_NUMBER: _ClassVar[int]
    REQUIRED_FIELD_NUMBER: _ClassVar[int]
    settingId: int
    settingTypeId: int
    settingTag: str
    settingName: str
    settingValue: str
    dataType: _ssocloud_pb2.DataType
    lastModified: str
    fromFile: bool
    encrypted: bool
    encoded: bool
    editable: bool
    translated: bool
    userVisible: bool
    required: bool
    def __init__(self, settingId: _Optional[int] = ..., settingTypeId: _Optional[int] = ..., settingTag: _Optional[str] = ..., settingName: _Optional[str] = ..., settingValue: _Optional[str] = ..., dataType: _Optional[_Union[_ssocloud_pb2.DataType, str]] = ..., lastModified: _Optional[str] = ..., fromFile: bool = ..., encrypted: bool = ..., encoded: bool = ..., editable: bool = ..., translated: bool = ..., userVisible: bool = ..., required: bool = ...) -> None: ...

class ApproveDeviceRequest(_message.Message):
    __slots__ = ["automatorId", "ssoAuthenticationProtocolType", "authMessage", "email", "devicePublicKey", "serverEccPublicKeyId", "userEncryptedDataKey", "userEncryptedDataKeyType", "ipAddress", "isTesting", "isEccOnly"]
    AUTOMATORID_FIELD_NUMBER: _ClassVar[int]
    SSOAUTHENTICATIONPROTOCOLTYPE_FIELD_NUMBER: _ClassVar[int]
    AUTHMESSAGE_FIELD_NUMBER: _ClassVar[int]
    EMAIL_FIELD_NUMBER: _ClassVar[int]
    DEVICEPUBLICKEY_FIELD_NUMBER: _ClassVar[int]
    SERVERECCPUBLICKEYID_FIELD_NUMBER: _ClassVar[int]
    USERENCRYPTEDDATAKEY_FIELD_NUMBER: _ClassVar[int]
    USERENCRYPTEDDATAKEYTYPE_FIELD_NUMBER: _ClassVar[int]
    IPADDRESS_FIELD_NUMBER: _ClassVar[int]
    ISTESTING_FIELD_NUMBER: _ClassVar[int]
    ISECCONLY_FIELD_NUMBER: _ClassVar[int]
    automatorId: int
    ssoAuthenticationProtocolType: SsoAuthenticationProtocolType
    authMessage: str
    email: str
    devicePublicKey: bytes
    serverEccPublicKeyId: int
    userEncryptedDataKey: bytes
    userEncryptedDataKeyType: _enterprise_pb2.EncryptedKeyType
    ipAddress: str
    isTesting: bool
    isEccOnly: bool
    def __init__(self, automatorId: _Optional[int] = ..., ssoAuthenticationProtocolType: _Optional[_Union[SsoAuthenticationProtocolType, str]] = ..., authMessage: _Optional[str] = ..., email: _Optional[str] = ..., devicePublicKey: _Optional[bytes] = ..., serverEccPublicKeyId: _Optional[int] = ..., userEncryptedDataKey: _Optional[bytes] = ..., userEncryptedDataKeyType: _Optional[_Union[_enterprise_pb2.EncryptedKeyType, str]] = ..., ipAddress: _Optional[str] = ..., isTesting: bool = ..., isEccOnly: bool = ...) -> None: ...

class SetupRequest(_message.Message):
    __slots__ = ["automatorId", "serverEccPublicKeyId", "automatorState", "encryptedEnterprisePrivateEccKey", "encryptedEnterprisePrivateRsaKey", "automatorSkills", "encryptedTreeKey", "isEccOnly"]
    AUTOMATORID_FIELD_NUMBER: _ClassVar[int]
    SERVERECCPUBLICKEYID_FIELD_NUMBER: _ClassVar[int]
    AUTOMATORSTATE_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTEDENTERPRISEPRIVATEECCKEY_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTEDENTERPRISEPRIVATERSAKEY_FIELD_NUMBER: _ClassVar[int]
    AUTOMATORSKILLS_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTEDTREEKEY_FIELD_NUMBER: _ClassVar[int]
    ISECCONLY_FIELD_NUMBER: _ClassVar[int]
    automatorId: int
    serverEccPublicKeyId: int
    automatorState: AutomatorState
    encryptedEnterprisePrivateEccKey: bytes
    encryptedEnterprisePrivateRsaKey: bytes
    automatorSkills: _containers.RepeatedCompositeFieldContainer[AutomatorSkill]
    encryptedTreeKey: bytes
    isEccOnly: bool
    def __init__(self, automatorId: _Optional[int] = ..., serverEccPublicKeyId: _Optional[int] = ..., automatorState: _Optional[_Union[AutomatorState, str]] = ..., encryptedEnterprisePrivateEccKey: _Optional[bytes] = ..., encryptedEnterprisePrivateRsaKey: _Optional[bytes] = ..., automatorSkills: _Optional[_Iterable[_Union[AutomatorSkill, _Mapping]]] = ..., encryptedTreeKey: _Optional[bytes] = ..., isEccOnly: bool = ...) -> None: ...

class StatusRequest(_message.Message):
    __slots__ = ["automatorId", "serverEccPublicKeyId", "isEccOnly"]
    AUTOMATORID_FIELD_NUMBER: _ClassVar[int]
    SERVERECCPUBLICKEYID_FIELD_NUMBER: _ClassVar[int]
    ISECCONLY_FIELD_NUMBER: _ClassVar[int]
    automatorId: int
    serverEccPublicKeyId: int
    isEccOnly: bool
    def __init__(self, automatorId: _Optional[int] = ..., serverEccPublicKeyId: _Optional[int] = ..., isEccOnly: bool = ...) -> None: ...

class InitializeRequest(_message.Message):
    __slots__ = ["automatorId", "idpMetadata", "idpSigningCertificate", "ssoEntityId", "emailMapping", "firstnameMapping", "lastnameMapping", "disabled", "serverEccPublicKeyId", "config", "sslMode", "persistState", "disableSniCheck", "sslCertificateFilename", "sslCertificateFilePassword", "sslCertificateKeyPassword", "sslCertificateContents", "automatorHost", "automatorPort", "ipAllow", "ipDeny", "isEccOnly"]
    AUTOMATORID_FIELD_NUMBER: _ClassVar[int]
    IDPMETADATA_FIELD_NUMBER: _ClassVar[int]
    IDPSIGNINGCERTIFICATE_FIELD_NUMBER: _ClassVar[int]
    SSOENTITYID_FIELD_NUMBER: _ClassVar[int]
    EMAILMAPPING_FIELD_NUMBER: _ClassVar[int]
    FIRSTNAMEMAPPING_FIELD_NUMBER: _ClassVar[int]
    LASTNAMEMAPPING_FIELD_NUMBER: _ClassVar[int]
    DISABLED_FIELD_NUMBER: _ClassVar[int]
    SERVERECCPUBLICKEYID_FIELD_NUMBER: _ClassVar[int]
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    SSLMODE_FIELD_NUMBER: _ClassVar[int]
    PERSISTSTATE_FIELD_NUMBER: _ClassVar[int]
    DISABLESNICHECK_FIELD_NUMBER: _ClassVar[int]
    SSLCERTIFICATEFILENAME_FIELD_NUMBER: _ClassVar[int]
    SSLCERTIFICATEFILEPASSWORD_FIELD_NUMBER: _ClassVar[int]
    SSLCERTIFICATEKEYPASSWORD_FIELD_NUMBER: _ClassVar[int]
    SSLCERTIFICATECONTENTS_FIELD_NUMBER: _ClassVar[int]
    AUTOMATORHOST_FIELD_NUMBER: _ClassVar[int]
    AUTOMATORPORT_FIELD_NUMBER: _ClassVar[int]
    IPALLOW_FIELD_NUMBER: _ClassVar[int]
    IPDENY_FIELD_NUMBER: _ClassVar[int]
    ISECCONLY_FIELD_NUMBER: _ClassVar[int]
    automatorId: int
    idpMetadata: str
    idpSigningCertificate: bytes
    ssoEntityId: str
    emailMapping: str
    firstnameMapping: str
    lastnameMapping: str
    disabled: bool
    serverEccPublicKeyId: int
    config: bytes
    sslMode: str
    persistState: bool
    disableSniCheck: bool
    sslCertificateFilename: str
    sslCertificateFilePassword: str
    sslCertificateKeyPassword: str
    sslCertificateContents: bytes
    automatorHost: str
    automatorPort: str
    ipAllow: str
    ipDeny: str
    isEccOnly: bool
    def __init__(self, automatorId: _Optional[int] = ..., idpMetadata: _Optional[str] = ..., idpSigningCertificate: _Optional[bytes] = ..., ssoEntityId: _Optional[str] = ..., emailMapping: _Optional[str] = ..., firstnameMapping: _Optional[str] = ..., lastnameMapping: _Optional[str] = ..., disabled: bool = ..., serverEccPublicKeyId: _Optional[int] = ..., config: _Optional[bytes] = ..., sslMode: _Optional[str] = ..., persistState: bool = ..., disableSniCheck: bool = ..., sslCertificateFilename: _Optional[str] = ..., sslCertificateFilePassword: _Optional[str] = ..., sslCertificateKeyPassword: _Optional[str] = ..., sslCertificateContents: _Optional[bytes] = ..., automatorHost: _Optional[str] = ..., automatorPort: _Optional[str] = ..., ipAllow: _Optional[str] = ..., ipDeny: _Optional[str] = ..., isEccOnly: bool = ...) -> None: ...

class NotInitializedResponse(_message.Message):
    __slots__ = ["automatorTransmissionKey", "signingCertificate", "signingCertificateFilename", "signingCertificatePassword", "signingKeyPassword", "signingCertificateFormat", "automatorPublicKey", "config"]
    AUTOMATORTRANSMISSIONKEY_FIELD_NUMBER: _ClassVar[int]
    SIGNINGCERTIFICATE_FIELD_NUMBER: _ClassVar[int]
    SIGNINGCERTIFICATEFILENAME_FIELD_NUMBER: _ClassVar[int]
    SIGNINGCERTIFICATEPASSWORD_FIELD_NUMBER: _ClassVar[int]
    SIGNINGKEYPASSWORD_FIELD_NUMBER: _ClassVar[int]
    SIGNINGCERTIFICATEFORMAT_FIELD_NUMBER: _ClassVar[int]
    AUTOMATORPUBLICKEY_FIELD_NUMBER: _ClassVar[int]
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    automatorTransmissionKey: bytes
    signingCertificate: bytes
    signingCertificateFilename: str
    signingCertificatePassword: str
    signingKeyPassword: str
    signingCertificateFormat: CertificateFormat
    automatorPublicKey: bytes
    config: bytes
    def __init__(self, automatorTransmissionKey: _Optional[bytes] = ..., signingCertificate: _Optional[bytes] = ..., signingCertificateFilename: _Optional[str] = ..., signingCertificatePassword: _Optional[str] = ..., signingKeyPassword: _Optional[str] = ..., signingCertificateFormat: _Optional[_Union[CertificateFormat, str]] = ..., automatorPublicKey: _Optional[bytes] = ..., config: _Optional[bytes] = ...) -> None: ...

class AutomatorResponse(_message.Message):
    __slots__ = ["automatorId", "enabled", "timestamp", "approveDevice", "status", "notInitialized", "error", "approveTeamsForUser", "approveTeams", "automatorState", "automatorPublicEccKey", "version"]
    AUTOMATORID_FIELD_NUMBER: _ClassVar[int]
    ENABLED_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    APPROVEDEVICE_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    NOTINITIALIZED_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    APPROVETEAMSFORUSER_FIELD_NUMBER: _ClassVar[int]
    APPROVETEAMS_FIELD_NUMBER: _ClassVar[int]
    AUTOMATORSTATE_FIELD_NUMBER: _ClassVar[int]
    AUTOMATORPUBLICECCKEY_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    automatorId: int
    enabled: bool
    timestamp: int
    approveDevice: ApproveDeviceResponse
    status: StatusResponse
    notInitialized: NotInitializedResponse
    error: ErrorResponse
    approveTeamsForUser: ApproveTeamsForUserResponse
    approveTeams: ApproveTeamsResponse
    automatorState: AutomatorState
    automatorPublicEccKey: bytes
    version: _version_pb2.Version
    def __init__(self, automatorId: _Optional[int] = ..., enabled: bool = ..., timestamp: _Optional[int] = ..., approveDevice: _Optional[_Union[ApproveDeviceResponse, _Mapping]] = ..., status: _Optional[_Union[StatusResponse, _Mapping]] = ..., notInitialized: _Optional[_Union[NotInitializedResponse, _Mapping]] = ..., error: _Optional[_Union[ErrorResponse, _Mapping]] = ..., approveTeamsForUser: _Optional[_Union[ApproveTeamsForUserResponse, _Mapping]] = ..., approveTeams: _Optional[_Union[ApproveTeamsResponse, _Mapping]] = ..., automatorState: _Optional[_Union[AutomatorState, str]] = ..., automatorPublicEccKey: _Optional[bytes] = ..., version: _Optional[_Union[_version_pb2.Version, _Mapping]] = ...) -> None: ...

class ApproveDeviceResponse(_message.Message):
    __slots__ = ["approved", "encryptedUserDataKey", "message", "encryptedUserDataKeyType"]
    APPROVED_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTEDUSERDATAKEY_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTEDUSERDATAKEYTYPE_FIELD_NUMBER: _ClassVar[int]
    approved: bool
    encryptedUserDataKey: bytes
    message: str
    encryptedUserDataKeyType: _enterprise_pb2.EncryptedKeyType
    def __init__(self, approved: bool = ..., encryptedUserDataKey: _Optional[bytes] = ..., message: _Optional[str] = ..., encryptedUserDataKeyType: _Optional[_Union[_enterprise_pb2.EncryptedKeyType, str]] = ...) -> None: ...

class StatusResponse(_message.Message):
    __slots__ = ["initialized", "enabledTimestamp", "initializedTimestamp", "updatedTimestamp", "numberOfDevicesApproved", "numberOfDevicesDenied", "numberOfErrors", "sslCertificateExpiration", "notInitializedResponse", "config", "numberOfTeamMembershipsApproved", "numberOfTeamMembershipsDenied", "numberOfTeamsApproved", "numberOfTeamsDenied"]
    INITIALIZED_FIELD_NUMBER: _ClassVar[int]
    ENABLEDTIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    INITIALIZEDTIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    UPDATEDTIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    NUMBEROFDEVICESAPPROVED_FIELD_NUMBER: _ClassVar[int]
    NUMBEROFDEVICESDENIED_FIELD_NUMBER: _ClassVar[int]
    NUMBEROFERRORS_FIELD_NUMBER: _ClassVar[int]
    SSLCERTIFICATEEXPIRATION_FIELD_NUMBER: _ClassVar[int]
    NOTINITIALIZEDRESPONSE_FIELD_NUMBER: _ClassVar[int]
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    NUMBEROFTEAMMEMBERSHIPSAPPROVED_FIELD_NUMBER: _ClassVar[int]
    NUMBEROFTEAMMEMBERSHIPSDENIED_FIELD_NUMBER: _ClassVar[int]
    NUMBEROFTEAMSAPPROVED_FIELD_NUMBER: _ClassVar[int]
    NUMBEROFTEAMSDENIED_FIELD_NUMBER: _ClassVar[int]
    initialized: bool
    enabledTimestamp: int
    initializedTimestamp: int
    updatedTimestamp: int
    numberOfDevicesApproved: int
    numberOfDevicesDenied: int
    numberOfErrors: int
    sslCertificateExpiration: int
    notInitializedResponse: NotInitializedResponse
    config: bytes
    numberOfTeamMembershipsApproved: int
    numberOfTeamMembershipsDenied: int
    numberOfTeamsApproved: int
    numberOfTeamsDenied: int
    def __init__(self, initialized: bool = ..., enabledTimestamp: _Optional[int] = ..., initializedTimestamp: _Optional[int] = ..., updatedTimestamp: _Optional[int] = ..., numberOfDevicesApproved: _Optional[int] = ..., numberOfDevicesDenied: _Optional[int] = ..., numberOfErrors: _Optional[int] = ..., sslCertificateExpiration: _Optional[int] = ..., notInitializedResponse: _Optional[_Union[NotInitializedResponse, _Mapping]] = ..., config: _Optional[bytes] = ..., numberOfTeamMembershipsApproved: _Optional[int] = ..., numberOfTeamMembershipsDenied: _Optional[int] = ..., numberOfTeamsApproved: _Optional[int] = ..., numberOfTeamsDenied: _Optional[int] = ...) -> None: ...

class ErrorResponse(_message.Message):
    __slots__ = ["message"]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    message: str
    def __init__(self, message: _Optional[str] = ...) -> None: ...

class LogEntry(_message.Message):
    __slots__ = ["serverTime", "messageLevel", "component", "message"]
    SERVERTIME_FIELD_NUMBER: _ClassVar[int]
    MESSAGELEVEL_FIELD_NUMBER: _ClassVar[int]
    COMPONENT_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    serverTime: str
    messageLevel: str
    component: str
    message: str
    def __init__(self, serverTime: _Optional[str] = ..., messageLevel: _Optional[str] = ..., component: _Optional[str] = ..., message: _Optional[str] = ...) -> None: ...

class AdminResponse(_message.Message):
    __slots__ = ["success", "message", "automatorInfo"]
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    AUTOMATORINFO_FIELD_NUMBER: _ClassVar[int]
    success: bool
    message: str
    automatorInfo: _containers.RepeatedCompositeFieldContainer[AutomatorInfo]
    def __init__(self, success: bool = ..., message: _Optional[str] = ..., automatorInfo: _Optional[_Iterable[_Union[AutomatorInfo, _Mapping]]] = ...) -> None: ...

class AutomatorInfo(_message.Message):
    __slots__ = ["automatorId", "nodeId", "name", "enabled", "url", "automatorSkills", "automatorSettingValues", "status", "logEntries", "automatorState", "version"]
    AUTOMATORID_FIELD_NUMBER: _ClassVar[int]
    NODEID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    ENABLED_FIELD_NUMBER: _ClassVar[int]
    URL_FIELD_NUMBER: _ClassVar[int]
    AUTOMATORSKILLS_FIELD_NUMBER: _ClassVar[int]
    AUTOMATORSETTINGVALUES_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    LOGENTRIES_FIELD_NUMBER: _ClassVar[int]
    AUTOMATORSTATE_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    automatorId: int
    nodeId: int
    name: str
    enabled: bool
    url: str
    automatorSkills: _containers.RepeatedCompositeFieldContainer[AutomatorSkill]
    automatorSettingValues: _containers.RepeatedCompositeFieldContainer[AutomatorSettingValue]
    status: StatusResponse
    logEntries: _containers.RepeatedCompositeFieldContainer[LogEntry]
    automatorState: AutomatorState
    version: str
    def __init__(self, automatorId: _Optional[int] = ..., nodeId: _Optional[int] = ..., name: _Optional[str] = ..., enabled: bool = ..., url: _Optional[str] = ..., automatorSkills: _Optional[_Iterable[_Union[AutomatorSkill, _Mapping]]] = ..., automatorSettingValues: _Optional[_Iterable[_Union[AutomatorSettingValue, _Mapping]]] = ..., status: _Optional[_Union[StatusResponse, _Mapping]] = ..., logEntries: _Optional[_Iterable[_Union[LogEntry, _Mapping]]] = ..., automatorState: _Optional[_Union[AutomatorState, str]] = ..., version: _Optional[str] = ...) -> None: ...

class AdminCreateAutomatorRequest(_message.Message):
    __slots__ = ["nodeId", "name", "skill"]
    NODEID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    SKILL_FIELD_NUMBER: _ClassVar[int]
    nodeId: int
    name: str
    skill: AutomatorSkill
    def __init__(self, nodeId: _Optional[int] = ..., name: _Optional[str] = ..., skill: _Optional[_Union[AutomatorSkill, _Mapping]] = ...) -> None: ...

class AdminDeleteAutomatorRequest(_message.Message):
    __slots__ = ["automatorId"]
    AUTOMATORID_FIELD_NUMBER: _ClassVar[int]
    automatorId: int
    def __init__(self, automatorId: _Optional[int] = ...) -> None: ...

class AdminGetAutomatorsOnNodeRequest(_message.Message):
    __slots__ = ["nodeId"]
    NODEID_FIELD_NUMBER: _ClassVar[int]
    nodeId: int
    def __init__(self, nodeId: _Optional[int] = ...) -> None: ...

class AdminGetAutomatorsForEnterpriseRequest(_message.Message):
    __slots__ = ["enterpriseId"]
    ENTERPRISEID_FIELD_NUMBER: _ClassVar[int]
    enterpriseId: int
    def __init__(self, enterpriseId: _Optional[int] = ...) -> None: ...

class AdminGetAutomatorRequest(_message.Message):
    __slots__ = ["automatorId"]
    AUTOMATORID_FIELD_NUMBER: _ClassVar[int]
    automatorId: int
    def __init__(self, automatorId: _Optional[int] = ...) -> None: ...

class AdminEnableAutomatorRequest(_message.Message):
    __slots__ = ["automatorId", "enabled"]
    AUTOMATORID_FIELD_NUMBER: _ClassVar[int]
    ENABLED_FIELD_NUMBER: _ClassVar[int]
    automatorId: int
    enabled: bool
    def __init__(self, automatorId: _Optional[int] = ..., enabled: bool = ...) -> None: ...

class AdminEditAutomatorRequest(_message.Message):
    __slots__ = ["automatorId", "name", "enabled", "url", "skillTypes", "automatorSettingValues"]
    AUTOMATORID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    ENABLED_FIELD_NUMBER: _ClassVar[int]
    URL_FIELD_NUMBER: _ClassVar[int]
    SKILLTYPES_FIELD_NUMBER: _ClassVar[int]
    AUTOMATORSETTINGVALUES_FIELD_NUMBER: _ClassVar[int]
    automatorId: int
    name: str
    enabled: bool
    url: str
    skillTypes: _containers.RepeatedScalarFieldContainer[SkillType]
    automatorSettingValues: _containers.RepeatedCompositeFieldContainer[AutomatorSettingValue]
    def __init__(self, automatorId: _Optional[int] = ..., name: _Optional[str] = ..., enabled: bool = ..., url: _Optional[str] = ..., skillTypes: _Optional[_Iterable[_Union[SkillType, str]]] = ..., automatorSettingValues: _Optional[_Iterable[_Union[AutomatorSettingValue, _Mapping]]] = ...) -> None: ...

class AdminSetupAutomatorRequest(_message.Message):
    __slots__ = ["automatorId", "automatorState", "encryptedEccEnterprisePrivateKey", "encryptedRsaEnterprisePrivateKey", "skillTypes", "encryptedTreeKey"]
    AUTOMATORID_FIELD_NUMBER: _ClassVar[int]
    AUTOMATORSTATE_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTEDECCENTERPRISEPRIVATEKEY_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTEDRSAENTERPRISEPRIVATEKEY_FIELD_NUMBER: _ClassVar[int]
    SKILLTYPES_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTEDTREEKEY_FIELD_NUMBER: _ClassVar[int]
    automatorId: int
    automatorState: AutomatorState
    encryptedEccEnterprisePrivateKey: bytes
    encryptedRsaEnterprisePrivateKey: bytes
    skillTypes: _containers.RepeatedScalarFieldContainer[SkillType]
    encryptedTreeKey: bytes
    def __init__(self, automatorId: _Optional[int] = ..., automatorState: _Optional[_Union[AutomatorState, str]] = ..., encryptedEccEnterprisePrivateKey: _Optional[bytes] = ..., encryptedRsaEnterprisePrivateKey: _Optional[bytes] = ..., skillTypes: _Optional[_Iterable[_Union[SkillType, str]]] = ..., encryptedTreeKey: _Optional[bytes] = ...) -> None: ...

class AdminSetupAutomatorResponse(_message.Message):
    __slots__ = ["success", "message", "automatorId", "automatorState", "automatorEccPublicKey"]
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    AUTOMATORID_FIELD_NUMBER: _ClassVar[int]
    AUTOMATORSTATE_FIELD_NUMBER: _ClassVar[int]
    AUTOMATORECCPUBLICKEY_FIELD_NUMBER: _ClassVar[int]
    success: bool
    message: str
    automatorId: int
    automatorState: AutomatorState
    automatorEccPublicKey: bytes
    def __init__(self, success: bool = ..., message: _Optional[str] = ..., automatorId: _Optional[int] = ..., automatorState: _Optional[_Union[AutomatorState, str]] = ..., automatorEccPublicKey: _Optional[bytes] = ...) -> None: ...

class AdminAutomatorSkillsRequest(_message.Message):
    __slots__ = ["automatorId"]
    AUTOMATORID_FIELD_NUMBER: _ClassVar[int]
    automatorId: int
    def __init__(self, automatorId: _Optional[int] = ...) -> None: ...

class AutomatorSkill(_message.Message):
    __slots__ = ["skillType", "name", "translatedName"]
    SKILLTYPE_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    TRANSLATEDNAME_FIELD_NUMBER: _ClassVar[int]
    skillType: SkillType
    name: str
    translatedName: str
    def __init__(self, skillType: _Optional[_Union[SkillType, str]] = ..., name: _Optional[str] = ..., translatedName: _Optional[str] = ...) -> None: ...

class AdminAutomatorSkillsResponse(_message.Message):
    __slots__ = ["success", "message", "automatorSkills"]
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    AUTOMATORSKILLS_FIELD_NUMBER: _ClassVar[int]
    success: bool
    message: str
    automatorSkills: _containers.RepeatedCompositeFieldContainer[AutomatorSkill]
    def __init__(self, success: bool = ..., message: _Optional[str] = ..., automatorSkills: _Optional[_Iterable[_Union[AutomatorSkill, _Mapping]]] = ...) -> None: ...

class AdminResetAutomatorRequest(_message.Message):
    __slots__ = ["automatorId"]
    AUTOMATORID_FIELD_NUMBER: _ClassVar[int]
    automatorId: int
    def __init__(self, automatorId: _Optional[int] = ...) -> None: ...

class AdminInitializeAutomatorRequest(_message.Message):
    __slots__ = ["automatorId"]
    AUTOMATORID_FIELD_NUMBER: _ClassVar[int]
    automatorId: int
    def __init__(self, automatorId: _Optional[int] = ...) -> None: ...

class AdminAutomatorLogRequest(_message.Message):
    __slots__ = ["automatorId"]
    AUTOMATORID_FIELD_NUMBER: _ClassVar[int]
    automatorId: int
    def __init__(self, automatorId: _Optional[int] = ...) -> None: ...

class AdminAutomatorLogClearRequest(_message.Message):
    __slots__ = ["automatorId"]
    AUTOMATORID_FIELD_NUMBER: _ClassVar[int]
    automatorId: int
    def __init__(self, automatorId: _Optional[int] = ...) -> None: ...

class ApproveTeamsForUserRequest(_message.Message):
    __slots__ = ["automatorId", "ssoAuthenticationProtocolType", "authMessage", "email", "serverEccPublicKeyId", "ipAddress", "userPublicKey", "teamDescription", "isTesting", "isEccOnly", "userPublicKeyEcc"]
    AUTOMATORID_FIELD_NUMBER: _ClassVar[int]
    SSOAUTHENTICATIONPROTOCOLTYPE_FIELD_NUMBER: _ClassVar[int]
    AUTHMESSAGE_FIELD_NUMBER: _ClassVar[int]
    EMAIL_FIELD_NUMBER: _ClassVar[int]
    SERVERECCPUBLICKEYID_FIELD_NUMBER: _ClassVar[int]
    IPADDRESS_FIELD_NUMBER: _ClassVar[int]
    USERPUBLICKEY_FIELD_NUMBER: _ClassVar[int]
    TEAMDESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    ISTESTING_FIELD_NUMBER: _ClassVar[int]
    ISECCONLY_FIELD_NUMBER: _ClassVar[int]
    USERPUBLICKEYECC_FIELD_NUMBER: _ClassVar[int]
    automatorId: int
    ssoAuthenticationProtocolType: SsoAuthenticationProtocolType
    authMessage: str
    email: str
    serverEccPublicKeyId: int
    ipAddress: str
    userPublicKey: bytes
    teamDescription: _containers.RepeatedCompositeFieldContainer[TeamDescription]
    isTesting: bool
    isEccOnly: bool
    userPublicKeyEcc: bytes
    def __init__(self, automatorId: _Optional[int] = ..., ssoAuthenticationProtocolType: _Optional[_Union[SsoAuthenticationProtocolType, str]] = ..., authMessage: _Optional[str] = ..., email: _Optional[str] = ..., serverEccPublicKeyId: _Optional[int] = ..., ipAddress: _Optional[str] = ..., userPublicKey: _Optional[bytes] = ..., teamDescription: _Optional[_Iterable[_Union[TeamDescription, _Mapping]]] = ..., isTesting: bool = ..., isEccOnly: bool = ..., userPublicKeyEcc: _Optional[bytes] = ...) -> None: ...

class TeamDescription(_message.Message):
    __slots__ = ["teamUid", "teamName", "encryptedTeamKey", "encryptedTeamKeyType"]
    TEAMUID_FIELD_NUMBER: _ClassVar[int]
    TEAMNAME_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTEDTEAMKEY_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTEDTEAMKEYTYPE_FIELD_NUMBER: _ClassVar[int]
    teamUid: bytes
    teamName: str
    encryptedTeamKey: bytes
    encryptedTeamKeyType: _enterprise_pb2.EncryptedKeyType
    def __init__(self, teamUid: _Optional[bytes] = ..., teamName: _Optional[str] = ..., encryptedTeamKey: _Optional[bytes] = ..., encryptedTeamKeyType: _Optional[_Union[_enterprise_pb2.EncryptedKeyType, str]] = ...) -> None: ...

class ApproveTeamsForUserResponse(_message.Message):
    __slots__ = ["automatorId", "email", "message", "approveTeamResponse"]
    AUTOMATORID_FIELD_NUMBER: _ClassVar[int]
    EMAIL_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    APPROVETEAMRESPONSE_FIELD_NUMBER: _ClassVar[int]
    automatorId: int
    email: str
    message: str
    approveTeamResponse: _containers.RepeatedCompositeFieldContainer[ApproveOneTeamForUserResponse]
    def __init__(self, automatorId: _Optional[int] = ..., email: _Optional[str] = ..., message: _Optional[str] = ..., approveTeamResponse: _Optional[_Iterable[_Union[ApproveOneTeamForUserResponse, _Mapping]]] = ...) -> None: ...

class ApproveOneTeamForUserResponse(_message.Message):
    __slots__ = ["approved", "message", "teamUid", "teamName", "userEncryptedTeamKey", "userEncryptedTeamKeyType", "userEncryptedTeamKeyByEcc", "userEncryptedTeamKeyByEccType"]
    APPROVED_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    TEAMUID_FIELD_NUMBER: _ClassVar[int]
    TEAMNAME_FIELD_NUMBER: _ClassVar[int]
    USERENCRYPTEDTEAMKEY_FIELD_NUMBER: _ClassVar[int]
    USERENCRYPTEDTEAMKEYTYPE_FIELD_NUMBER: _ClassVar[int]
    USERENCRYPTEDTEAMKEYBYECC_FIELD_NUMBER: _ClassVar[int]
    USERENCRYPTEDTEAMKEYBYECCTYPE_FIELD_NUMBER: _ClassVar[int]
    approved: bool
    message: str
    teamUid: bytes
    teamName: str
    userEncryptedTeamKey: bytes
    userEncryptedTeamKeyType: _enterprise_pb2.EncryptedKeyType
    userEncryptedTeamKeyByEcc: bytes
    userEncryptedTeamKeyByEccType: _enterprise_pb2.EncryptedKeyType
    def __init__(self, approved: bool = ..., message: _Optional[str] = ..., teamUid: _Optional[bytes] = ..., teamName: _Optional[str] = ..., userEncryptedTeamKey: _Optional[bytes] = ..., userEncryptedTeamKeyType: _Optional[_Union[_enterprise_pb2.EncryptedKeyType, str]] = ..., userEncryptedTeamKeyByEcc: _Optional[bytes] = ..., userEncryptedTeamKeyByEccType: _Optional[_Union[_enterprise_pb2.EncryptedKeyType, str]] = ...) -> None: ...

class ApproveTeamsRequest(_message.Message):
    __slots__ = ["automatorId", "ssoAuthenticationProtocolType", "authMessage", "email", "serverEccPublicKeyId", "ipAddress", "teamDescription", "isEccOnly", "isTesting"]
    AUTOMATORID_FIELD_NUMBER: _ClassVar[int]
    SSOAUTHENTICATIONPROTOCOLTYPE_FIELD_NUMBER: _ClassVar[int]
    AUTHMESSAGE_FIELD_NUMBER: _ClassVar[int]
    EMAIL_FIELD_NUMBER: _ClassVar[int]
    SERVERECCPUBLICKEYID_FIELD_NUMBER: _ClassVar[int]
    IPADDRESS_FIELD_NUMBER: _ClassVar[int]
    TEAMDESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    ISECCONLY_FIELD_NUMBER: _ClassVar[int]
    ISTESTING_FIELD_NUMBER: _ClassVar[int]
    automatorId: int
    ssoAuthenticationProtocolType: SsoAuthenticationProtocolType
    authMessage: str
    email: str
    serverEccPublicKeyId: int
    ipAddress: str
    teamDescription: _containers.RepeatedCompositeFieldContainer[TeamDescription]
    isEccOnly: bool
    isTesting: bool
    def __init__(self, automatorId: _Optional[int] = ..., ssoAuthenticationProtocolType: _Optional[_Union[SsoAuthenticationProtocolType, str]] = ..., authMessage: _Optional[str] = ..., email: _Optional[str] = ..., serverEccPublicKeyId: _Optional[int] = ..., ipAddress: _Optional[str] = ..., teamDescription: _Optional[_Iterable[_Union[TeamDescription, _Mapping]]] = ..., isEccOnly: bool = ..., isTesting: bool = ...) -> None: ...

class ApproveTeamsResponse(_message.Message):
    __slots__ = ["automatorId", "message", "approveTeamResponse"]
    AUTOMATORID_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    APPROVETEAMRESPONSE_FIELD_NUMBER: _ClassVar[int]
    automatorId: int
    message: str
    approveTeamResponse: _containers.RepeatedCompositeFieldContainer[ApproveOneTeamResponse]
    def __init__(self, automatorId: _Optional[int] = ..., message: _Optional[str] = ..., approveTeamResponse: _Optional[_Iterable[_Union[ApproveOneTeamResponse, _Mapping]]] = ...) -> None: ...

class ApproveOneTeamResponse(_message.Message):
    __slots__ = ["approved", "message", "teamUid", "teamName", "encryptedTeamKeyCbc", "encryptedTeamKeyCbcType", "encryptedTeamKeyGcm", "encryptedTeamKeyGcmType", "teamPublicKeyRsa", "encryptedTeamPrivateKeyRsa", "encryptedTeamPrivateKeyRsaType", "teamPublicKeyEcc", "encryptedTeamPrivateKeyEcc", "encryptedTeamPrivateKeyEccType"]
    APPROVED_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    TEAMUID_FIELD_NUMBER: _ClassVar[int]
    TEAMNAME_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTEDTEAMKEYCBC_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTEDTEAMKEYCBCTYPE_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTEDTEAMKEYGCM_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTEDTEAMKEYGCMTYPE_FIELD_NUMBER: _ClassVar[int]
    TEAMPUBLICKEYRSA_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTEDTEAMPRIVATEKEYRSA_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTEDTEAMPRIVATEKEYRSATYPE_FIELD_NUMBER: _ClassVar[int]
    TEAMPUBLICKEYECC_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTEDTEAMPRIVATEKEYECC_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTEDTEAMPRIVATEKEYECCTYPE_FIELD_NUMBER: _ClassVar[int]
    approved: bool
    message: str
    teamUid: bytes
    teamName: str
    encryptedTeamKeyCbc: bytes
    encryptedTeamKeyCbcType: _enterprise_pb2.EncryptedKeyType
    encryptedTeamKeyGcm: bytes
    encryptedTeamKeyGcmType: _enterprise_pb2.EncryptedKeyType
    teamPublicKeyRsa: bytes
    encryptedTeamPrivateKeyRsa: bytes
    encryptedTeamPrivateKeyRsaType: _enterprise_pb2.EncryptedKeyType
    teamPublicKeyEcc: bytes
    encryptedTeamPrivateKeyEcc: bytes
    encryptedTeamPrivateKeyEccType: _enterprise_pb2.EncryptedKeyType
    def __init__(self, approved: bool = ..., message: _Optional[str] = ..., teamUid: _Optional[bytes] = ..., teamName: _Optional[str] = ..., encryptedTeamKeyCbc: _Optional[bytes] = ..., encryptedTeamKeyCbcType: _Optional[_Union[_enterprise_pb2.EncryptedKeyType, str]] = ..., encryptedTeamKeyGcm: _Optional[bytes] = ..., encryptedTeamKeyGcmType: _Optional[_Union[_enterprise_pb2.EncryptedKeyType, str]] = ..., teamPublicKeyRsa: _Optional[bytes] = ..., encryptedTeamPrivateKeyRsa: _Optional[bytes] = ..., encryptedTeamPrivateKeyRsaType: _Optional[_Union[_enterprise_pb2.EncryptedKeyType, str]] = ..., teamPublicKeyEcc: _Optional[bytes] = ..., encryptedTeamPrivateKeyEcc: _Optional[bytes] = ..., encryptedTeamPrivateKeyEccType: _Optional[_Union[_enterprise_pb2.EncryptedKeyType, str]] = ...) -> None: ...
